% W-H system simulation program
% nonlinear system y = av1*x + av2*x^2 + av3*x^3 + av4*x^4+ av5*x^5,
%
% x - L - x1 - NL - x2 - R -y
%
clear all; clf; rand('seed',0);
%
av=[1 .1*[0.2 0.5 0.05 0.05]];
% 'irg' - ideal gaussian, 'rms' - multisine, 

ML=100;                   %number of measurements
npoints=1024*2*4; 		% number of time domain points
n=npoints/2;     		% maximum number of frequencies
nfreq=fix(npoints/10);
junit=sqrt(-1);
%
% frequency axes
fv=[1:nfreq]';
fvrel=0.1*fv/nfreq; zet=exp(-sqrt(-1)*2*pi*fvrel);
%
% W-H system design, 'L' input dynamics, 'R' output dynamics
[b1,a1]=butter(3,[0.03 0.07]); % on 0 ... 0.1 axis 3dB points are 0.015 0.035
[b1,a1]=butter(0,0.1);
[b2,a2]=butter(3,[0.03 0.07]);
%
Lb1=length(b1); HLbf=0; for i=1:Lb1, HLbf=HLbf+b1(i)*zet.^(i-1); end;
La1=length(a1); HLaf=0; for i=1:La1, HLaf=HLaf+a1(i)*zet.^(i-1); end;
HLf=HLbf./HLaf;
%
Lb2=length(b2); HRbf=0; for i=1:Lb2, HRbf=HRbf+b2(i)*zet.^(i-1); end;
La2=length(a2); HRaf=0; for i=1:La2, HRaf=HRaf+a2(i)*zet.^(i-1); end; 
HRf=HRbf./HRaf;
%
LRf=HLf.*HRf;
%
% BLA dynamic systems
RDS1=1;
RDS3=3*sum(abs(HLf(fv)).^2)/length(fv);
RDS5=5*3*(sum(abs(HLf(fv)).^2)/length(fv))^2;
RDStr=av(1)*RDS1+av(3)*RDS3+av(5)*RDS5;
%
figure(1),
 subplot(2,2,1), plot(fvrel,db(HLf)); grid; axis([0 .1 -50 10]);
 subplot(2,2,2), plot(fvrel,db(HRf)); grid; axis([0 .1 -50 10]);
 subplot(2,2,4), plot(fvrel,db(LRf)); grid; axis([0 .1 -50 10]);
 title('L(f), R(f), LR(f)');
 pause;
%
Xrms=zeros(npoints,1); xrms=Xrms;
Xirg=zeros(npoints,1); xirg=Xrms;
%
Grms=zeros(length(fv),1); 
Girg=zeros(length(fv),1); 
Syxirg=Girg; Sxxirg=Girg; 
GrmsAv=0*Grms; GirgAv=0*Girg; 

VG=zeros(length(fv),1);
%
for ll=1:ML, ll
% xrms, Xrms - random phase multisine
 fi=rand(length(fv),1)*2*pi; Zx=exp(-junit*fi); Xrms=0*Xrms;
 Xrms(fv+1)=Zx; Xrms(2*n+1-fv)=conj(Xrms(1+fv));
 xrms=real(ifft(Xrms)); Ax=sqrt(sum(xrms.^2)/npoints);  xrms=xrms/Ax;
 Xrms=fft(xrms); AW=abs(Xrms)/max(abs(Xrms));

% xirg, Xirg - ideal filtered Gaussian signal
 xaux=randn(length(xrms),1); Xaux=fft(xaux); Xirg=Xaux.*AW;
 xirg=real(ifft(Xirg)); Ax=sqrt(sum(xirg.^2)/npoints); xirg=xirg/Ax;
 Xirg=fft(xirg);

xrmsfull=[xrms; xrms; xrms; xrms];              x1rms=filter(b1,a1,xrmsfull);
xrmsfull=[xirg; xirg; xirg; xirg];              x1irg=filter(b1,a1,xrmsfull);

Nfull=length(xrmsfull);
Nfullminrec=Nfull-npoints+1;

% passing signals through nonlinear system
x2rms=0*x1rms; x2irg=0*x1rms; 

if av(1)~=0,
  x2rms=x2rms+av(1)*x1rms; x2irg=x2irg+av(1)*x1irg;
end;
if av(2)~=0,
  x2rms=x2rms+av(2)*x1rms.^2; x2irg=x2irg+av(2)*x1irg.^2; 
end;
if av(3)~=0,
  x2rms=x2rms+av(3)*x1rms.^3; x2irg=x2irg+av(3)*x1irg.^3;
end;
if av(4)~=0,
  x2rms=x2rms+av(4)*x1rms.^4; x2irg=x2irg+av(4)*x1irg.^4;
end;
if av(5)~=0,
  x2rms=x2rms+av(5)*x1rms.^5; x2irg=x2irg+av(5)*x1irg.^5; 
end;
%
% passing signals through output dynamics
%
yrms=filter(b2,a2,x2rms);       Yrms=fft(yrms(Nfullminrec:Nfull));
yirg=filter(b2,a2,x2irg);       Yirg=fft(yirg(Nfullminrec:Nfull));
%
% computing estimates of the system transfer characteristics
 Grms=Yrms(fv+1)./Xrms(fv+1);
 
 GrmsAv=GrmsAv+(Grms-GrmsAv)/ll;
 
 VG=VG+((abs(Grms-GrmsAv)).^2-VG)/ll;

 % for gaussian signals
 Syxirg=Yirg(fv+1).*conj(Xirg(fv+1));
 Sxxirg=Xirg(fv+1).*conj(Xirg(fv+1));
 Girg=Syxirg./Sxxirg;
 
 GirgAv=GirgAv+(Girg-GirgAv)/ll;

strcol='c';
 figure(11),
 subplot(1,2,2), plot(fv,db(Grms(:)),strcol,fv,db(RDStr*LRf),'r',fv,db(GrmsAv(:)),'b');
 axis([0 1000 -30 20]); grid; title('RMS','Fontsize',14);
 subplot(1,2,1), plot(fv,db(Girg(:)),strcol,fv,db(RDStr*LRf),'r',fv,db(GirgAv(:)),'b');
 axis([0 1000 -30 20]); grid; title('Gauss','Fontsize',14);
 drawnow;
%pause;
end;

figure(12),
subplot(1,2,2), plot(fv,db(RDStr*LRf),'r',fv,db(GrmsAv(:)),'b');
axis([0 1000 -30 20]); grid; title('RMS','Fontsize',14);
subplot(1,2,1), plot(fv,db(RDStr*LRf),'r',fv,db(GirgAv(:)),'b');
axis([0 1000 -30 20]); grid; title('Gauss','Fontsize',14);

xv=[-3:.01:3];
yv=av(1)*xv+av(2)*xv.^2+av(3)*xv.^3+av(4)*xv.^4+av(5)*xv.^5;
figure(16), plot(xv,yv,'o'); grid;
title('Nonlinearity','Fontsize',14);


